from pyqtgraph.Qt import QtWidgets
from functools import wraps
import math

class Piece(QtWidgets.QGroupBox):
    """
    A single `Piece` object is an unit of automation - an object that is meant to represent a single
    physical instrument (like a laser) or a particular functionality (like a plotter or a parameter scan).

    Pieces can be assembled into a :class:`~puzzlepiece.puzzle.Puzzle`.

    :param puzzle: The parent :class:`~puzzlepiece.puzzle.Puzzle`.
    :param custom_horizontal: A bool flat, the custom layout is displayed to the right of the main controls
                                if True.
    """
    def __init__(self, puzzle, custom_horizontal=False, *args, **kwargs):
        super().__init__()
        #: Reference to the parent :class:`~puzzlepiece.puzzle.Puzzle`.
        self.puzzle = puzzle
        #: Boolean flag. See :func:`~puzzlepiece.piece.Piece.call_stop`
        self.stop = False

        #: dict: A dictionary of this Piece's params (see :class:`~puzzlepiece.param.BaseParam`)
        self.params = {}
        # A reference to the param dictionary for backwards-compatibility
        self.readouts = self.params
        #: dict: A dictionary of this Piece's actions (see :class:`~puzzlepiece.action.Action`)
        self.actions = {}
        self.shortcuts = {}

        if not self.puzzle.debug:
            self.setup()
        
        self.define_params()
        self.define_readouts()
        self.define_actions()
        self.folder = None

        if custom_horizontal:
            self.layout = QtWidgets.QHBoxLayout()
        else:
            self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        control_layout = QtWidgets.QVBoxLayout()
        control_layout.addLayout(self.param_layout())
        control_layout.addLayout(self.action_layout())
        self.layout.addLayout(control_layout)

        custom_layout = self.custom_layout()
        if custom_layout is not None:
            self.layout.addLayout(custom_layout)

        if custom_layout is None or custom_horizontal:
            control_layout.addStretch()

    def param_layout(self, wrap=1):
        """
        Genereates a `QGridLayout` for the params. Override to set a different wrapping.

        :param wrap: the number of columns the params are displayed in .
        :rtype: QtWidgets.QGridLayout
        """
        layout = QtWidgets.QGridLayout()
        visible_params = [key for key in self.params if self.params[key].visible]
        numrows = math.ceil(len(visible_params) / wrap)
        for i, key in enumerate(visible_params):
            layout.addWidget(self.params[key], i%numrows, i//numrows)
        return layout

    def action_layout(self, wrap=2):
        """
        Genereates a `QGridLayout` for the actions. Override to set a different wrapping.

        :param wrap: the number of columns the actions are displayed in .
        :rtype: QtWidgets.QGridLayout
        """
        layout = QtWidgets.QGridLayout()
        visible_actions = [key for key in self.actions if self.actions[key].visible]
        for i, key in enumerate(visible_actions):
            button = QtWidgets.QPushButton(key)
            button.clicked.connect(lambda x=False, _key=key: self.actions[_key]())
            layout.addWidget(button, i//wrap, i%wrap)
        return layout

    def custom_layout(self):
        """
        Override to generate a custom `QLayout` that this Piece will display.

        :rtype: QtWidgets.QLayout
        """
        return None

    def define_params(self):
        """
        Override to define params using decorators from :mod:`puzzlepiece.param`.
        """
        pass

    def define_readouts(self):
        """
        Override to define readouts (params with getters). This is no different that defining them in
        :func:`~puzzlepiece.piece.Piece.define_params`, but may be a convenient way to organise the
        definitions within your custom class.
        """
        pass

    def define_actions(self):
        """
        Override to define actions using decorators from :mod:`puzzlepiece.action`.
        """
        pass

    def setup(self):
        """
        Only called if the :class:`~puzzlepiece.puzzle.Puzzle` debug flag is False.
        Override to set up necessary hardware libraries.
        """
        pass

    def call_stop(self):
        """
        This method is called by the parent Puzzle when a global stop is called.

        By default, it sets the stop flag to True. Detect the flag in you code to stop processes.

        Alternatively, this can be overriden to support more complex actions.
        """
        self.stop = True

    def handle_close(self, event):
        """
        Only called if the :class:`~puzzlepiece.puzzle.Puzzle` debug flag is False.
        Override to disconnect hardware etc when the main window closes.
        """
        pass

    def handle_shortcut(self, event):
        """
        Calls an Action if a keyboard shortcut has been registered.

        :meta private:
        """
        if event.key() in self.shortcuts:
            self.shortcuts[event.key()]()

    def elevate(self):
        """
        If this Piece resides in a :class:`~puzzlepiece.puzzle.Folder`, this method switches the tab
        to make this Piece visible.
        """
        if self.folder is not None:
            self.folder.setCurrentWidget(self)


def ensurer(ensure_function):
    """
    An ensurer is a decorator that can be placed on getters, setters, and actions, and it will run
    ahead of these functions. The intended behaviour is performing checks ahead of running the
    function - for example checking if a laser is connected ahead of trying to set its power.
    This way one ensurer can be written and used in multiple places easily.

    Note that **the method being decorated should raise an exception if the check fails!** This way
    execution will stop if the condition is not met. This is not mandatory though - custom behaviour
    is allowed.

    For example, an ensurer can be defined as a Piece's method (in the main body of the class).::

        @puzzlepiece.piece.ensurer
        def _ensure_connected(self):
            if not self.params['connected'].get_value():
                raise Exception('Laser not connected') 
    
    This can then be used when defining a param (below the param-defining decorator)::

        @puzzlepiece.param.spinbox(self, 'power', 0.)
        @self._ensure_connected
        def power(self, value):
            self.laser.set_power(value)

        @puzzlepiece.param.spinbox(self, 'wavelength', 0.)
        @self._ensure_connected
        def wavelength(self, value):
            self.laser.set_wavelength(value)

    It can also be called directly if preferred, optionally with `capture_exception=True`
    which will return True if the check passes, or False if the check raises an Exception::

        # This should raise an Exception is the check fails
        self._ensure_connected()

        # This will not raise an Exception is the check fails
        if self._ensure_connected(capture_exception=True):
            print("laser is connected!")
    """
    # Decorating a class method with ensure makes it a decorator.
    # Here we create this decorator and return it. 
    @wraps(ensure_function)
    def ensure_decorator(self, main_function=None, capture_exception=False):
        if main_function is not None:
            # This means ensure_decorator was used as a decorator, and
            # main_function is the function being decorated. We therefore
            # wrap it with the ensuring functionality and return it 
            @wraps(main_function)
            def wrapped_main(self, *args, **kwargs):
                ensure_function(self)
                return main_function(self, *args, **kwargs)
            return wrapped_main
        else:
            # If main_function is None, ensure_decorator has been called
            # directly instead of being used as a decorator, so we 
            # just execute ensure_function
            if capture_exception:
                try:
                    ensure_function(self)
                except:
                    return False
                return True
            else:
                ensure_function(self)
    return ensure_decorator


from qtpy import QtCore, QtWidgets
import time

class _Emitter(QtCore.QObject):
    # The Emitter is needed as a QRunnable is not a QObject, and cannot emit it's own signals.
    # So we set up the Signal here, and let a Worker instance an Emitter for its use
    signal = QtCore.Signal(object)

class Worker(QtCore.QRunnable):
    """
    Generic worker (QRunnable) that calls a function in a thread.

    :param function: The function for the Worker to run.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """
    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.done = False
        self._emitter = _Emitter()
        #: A Qt signal emitted when the function returns, passes the returned value to the connected Slot.
        self.returned = self._emitter.signal
        super().__init__()

    @QtCore.Slot()
    def run(self):
        """
        Run the Worker. Shouldn't be excecuted directly, instead
        use :func:`puzzlepiece.puzzle.Puzzle.run_worker`.
        """
        try:
            r = self.function(*self.args, **self.kwargs)
            self.returned.emit(r)
        finally:
            self.done=True


class LiveWorker(Worker):
    """
    A Worker that calls a function repeatedly in a thread,
    separated by specified time intervals.

    :param function: The function for the Worker to run.
    :param sleep: Time to sleep between function calls in seconds.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """
    def __init__(self, function, sleep=.1, args=None, kwargs=None):
        self.stopping = False
        self.sleep = sleep
        super().__init__(function, args, kwargs)
        #: A Qt signal emitted each time the function returns, passes the returned value to the connected Slot.
        self.returned = self.returned

    def stop(self):
        """
        Ask the Worker to stop. This will only take effect once the current
        execution of the function is over.
        """
        self.stopping = True

    @QtCore.Slot()
    def run(self):
        """
        Run the Worker. Shouldn't be excecuted directly, instead
        use :func:`puzzlepiece.puzzle.Puzzle.run_worker`.
        """
        try:
            while not self.stopping:
                r = self.function(*self.args, **self.kwargs)
                self.returned.emit(r)
                if not self.stopping:
                    time.sleep(self.sleep)
        finally:
            self.done = True


class PuzzleTimer(QtWidgets.QWidget):
    """
    A Widget that displays a checkbox. When the checkbox is checked, an associated
    function is called repeatedly in a thread, separated by specified time intervals.

    :param name: The display name for this PuzzleTimer.
    :param puzzle: The parent :class:`~puzzlepiece.puzzle.Puzzle` (needed as the Puzzle
      is charged with the QThreadPool that runs the tasks).
    :param function: The function for the PuzzleTimer to run.
    :param sleep: Time to sleep between function calls in seconds.
    :param args: list of arguments to be forwarded to the function when run.
    :param kwargs: dictionary of keyword arguments to be forwarded to the function when run.
    """
    #: A Qt signal emitted each time the associated LiveWorker returns, passes the returned value to the connected Slot.
    returned = QtCore.Signal(object)
    def __init__(self, name, puzzle, function, sleep=.1, args=None, kwargs=None):
        self.name = name
        self.puzzle = puzzle
        self.function = function
        self._sleep = sleep
        self.args = args
        self.kwargs = kwargs
        self.worker = None

        super().__init__()

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0,0,0,0)
        self.setLayout(layout)

        self.input = QtWidgets.QCheckBox(self.name)
        layout.addWidget(self.input)
        self.input.stateChanged.connect(self._state_handler)

    def _state_handler(self, state):
        if state and (self.worker is None or self.worker.done):
            self.worker = LiveWorker(self.function, self._sleep, self.args, self.kwargs)
            self.worker.returned.connect(self._return_handler)
            self.puzzle.run_worker(self.worker)
        elif self.worker is not None:
            self.worker.stop()

    def _return_handler(self, value):
        self.returned.emit(value)

    def stop(self):
        """
        Ask the PuzzleTimer to stop.
        
        This will only take effect once the current execution of the function is over,
        see :func:`puzzlepiece.threads.LiveWorker.stop`.
        """
        self.input.setChecked(False)

    @property
    def sleep(self):
        """
        Interval in seconds between executions of the associated function.
        The property can be modified to change the interval for the associated Worker,
        even if it's already running.
        """
        return self._sleep
    
    @sleep.setter
    def sleep(self, value):
        if self.worker is not None:
            self.worker.sleep = float(value)
        self._sleep = float(value)


from . import piece
Piece = piece.Piece
from . import puzzle
Puzzle = puzzle.Puzzle
QApp = puzzle.QApp

from . import param
from . import readout
from . import action
from . import parse
from . import threads


import puzzlepiece as pzp
from pyqtgraph.Qt import QtWidgets

class TParamPiece(pzp.Piece):
    def define_params(self):
        pzp.param.base_param(self, 'base_param', 0)(None)

        self._setter_param_value = None
        @pzp.param.base_param(self, 'setter_param', 1)
        def setter_param(self, value):
            self._setter_param_value = value

        @pzp.param.base_param(self, 'setter_return_param', 0)
        def setter_param(self, value):
            return value + 1

        @pzp.param.readout(self, 'getter_param', _type=int)
        def getter_param(self):
            return 1

        self._setter_getter_param_value = 0
        @pzp.param.base_param(self, 'setter_getter_param', 0)
        def setter_getter_param(self, value):
            self._setter_getter_param_value = value
        
        @setter_getter_param.set_getter(self)
        def setter_getter_param_getter(self):
            return self._setter_getter_param_value + 1
        
        self._setter_getter_return_param_value = 0
        @pzp.param.base_param(self, 'setter_getter_return_param', 0)
        def setter_getter_return_param(self, value):
            self._setter_getter_return_param_value = value
            return value
        
        @setter_getter_return_param.set_getter(self)
        def setter_getter_return_param_getter(self):
            return self._setter_getter_return_param_value + 1

        @pzp.param.base_param(self, 'error_param', 0)
        def error_param(self, value):
            # return 0
            raise Exception('Setter exception')
        
        @error_param.set_getter(self)
        def error_param_getter(self):
            # return 0
            raise Exception('Getter exception')

        @pzp.param.checkbox(self, 'error_checkbox', 0)
        def error_param(self, value):
            raise Exception('Setter exception')
        
        pzp.param.base_param(self, 'float_param', 0.)(None)
        pzp.param.base_param(self, 'format_param', 0., format='{:.2f}')(None)
        pzp.param.spinbox(self, 'input_param', 0.)(None)
        pzp.param.text(self, 'text_param', "")(None)
        

def test_base_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]
    def count_calls(count=count):
        count[0] += 1
    puzzle['test'].params['base_param'].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle['test'].params['base_param'].value == 0
    assert puzzle['test'].params['base_param'].get_value() == 0
    assert puzzle['test'].params['base_param'].input.text() == '0'
    assert count[0] == 0

    puzzle['test'].params['base_param'].set_value(1)
    assert count[0] == 1
    assert puzzle['test'].params['base_param'].value == 1
    assert puzzle['test'].params['base_param'].get_value() == 1
    assert puzzle['test'].params['base_param'].input.text() == '1'

def test_setter_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]
    def count_calls(count=count):
        count[0] += 1
    puzzle['test'].params['setter_param'].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle['test']._setter_param_value is None
    assert puzzle['test'].params['setter_param'].value is None
    assert puzzle['test'].params['setter_param'].get_value() is None
    assert puzzle['test'].params['setter_param'].input.text() == '1'
    assert count[0] == 0

    puzzle['test'].params['setter_param'].set_value()
    assert count[0] == 1
    assert puzzle['test']._setter_param_value == 1
    assert puzzle['test'].params['setter_param'].value == 1
    assert puzzle['test'].params['setter_param'].get_value() == 1
    assert puzzle['test'].params['setter_param'].input.text() == '1'

    puzzle['test'].params['setter_param'].set_value(2)
    assert count[0] == 2
    assert puzzle['test']._setter_param_value == 2
    assert puzzle['test'].params['setter_param'].value == 2
    assert puzzle['test'].params['setter_param'].get_value() == 2
    assert puzzle['test'].params['setter_param'].input.text() == '2'

    puzzle['test'].params['setter_param']._input_set_value(3)
    assert count[0] == 2
    assert puzzle['test'].params['setter_param'].input.text() == '3'
    puzzle['test'].params['setter_param']._set_button.click()
    assert count[0] == 3
    assert puzzle['test']._setter_param_value == 3
    assert puzzle['test'].params['setter_param'].value == 3
    assert puzzle['test'].params['setter_param'].get_value() == 3
    assert puzzle['test'].params['setter_param'].input.text() == '3'

def test_setter_return_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]
    def count_calls(count=count):
        count[0] += 1
    puzzle['test'].params['setter_return_param'].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle['test'].params['setter_return_param'].value is None
    assert puzzle['test'].params['setter_return_param'].get_value() is None
    assert puzzle['test'].params['setter_return_param'].input.text() == '0'
    assert count[0] == 0

    puzzle['test'].params['setter_return_param'].set_value()
    assert count[0] == 1
    assert puzzle['test'].params['setter_return_param'].value == 1
    assert puzzle['test'].params['setter_return_param'].get_value() == 1
    assert puzzle['test'].params['setter_return_param'].input.text() == '1'

    puzzle['test'].params['setter_return_param'].set_value(2)
    assert count[0] == 2
    assert puzzle['test'].params['setter_return_param'].value == 3
    assert puzzle['test'].params['setter_return_param'].get_value() == 3
    assert puzzle['test'].params['setter_return_param'].input.text() == '3'

    puzzle['test'].params['setter_return_param']._input_set_value(4)
    assert count[0] == 2
    assert puzzle['test'].params['setter_return_param'].input.text() == '4'
    puzzle['test'].params['setter_return_param']._set_button.click()
    assert count[0] == 3
    assert puzzle['test'].params['setter_return_param'].value == 5
    assert puzzle['test'].params['setter_return_param'].get_value() == 5
    assert puzzle['test'].params['setter_return_param'].input.text() == '5'

def test_getter_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]
    def count_calls(count=count):
        count[0] += 1
    puzzle['test'].params['getter_param'].changed.connect(count_calls)

    assert count[0] == 0
    assert puzzle['test'].params['getter_param'].value is None
    assert puzzle['test'].params['getter_param'].input.text() == ''
    assert count[0] == 0

    puzzle['test'].params['getter_param'].get_value()
    assert count[0] == 1
    assert puzzle['test'].params['getter_param'].value == 1
    assert puzzle['test'].params['getter_param'].input.text() == '1'

    puzzle['test'].params['getter_param'].get_value()
    assert count[0] == 2
    assert puzzle['test'].params['getter_param'].value == 1
    assert puzzle['test'].params['getter_param'].input.text() == '1'

    # Updated design: set_value in general emits the signal,
    # even if the input field is unchanged
    puzzle['test'].params['getter_param'].set_value()
    assert count[0] == 3
    assert puzzle['test'].params['getter_param'].value == 1
    assert puzzle['test'].params['getter_param'].input.text() == '1'

    # In contrast, changing the value of the input with the hidden function _input_set_value
    # should in general not emit the signal
    puzzle['test'].params['getter_param']._input_set_value(2)
    assert count[0] == 3
    assert puzzle['test'].params['getter_param'].value == 1
    assert puzzle['test'].params['getter_param'].input.text() == '2'

    puzzle['test'].params['getter_param'].set_value(3)
    assert count[0] == 4
    assert puzzle['test'].params['getter_param'].value == 3
    assert puzzle['test'].params['getter_param'].input.text() == '3'

    puzzle['test'].params['getter_param'].get_value()
    assert count[0] == 5
    assert puzzle['test'].params['getter_param'].value == 1
    assert puzzle['test'].params['getter_param'].input.text() == '1'

def test_precision_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    # The base_param has an int format, so it should cast values given to int
    puzzle['test'].params['base_param'].set_value(0.1234)
    assert puzzle['test'].params['base_param'].value == 0
    assert puzzle['test'].params['base_param'].get_value() == 0

    # The float_param has a float format, so it should retain precision
    puzzle['test'].params['float_param'].set_value(0.1234)
    assert puzzle['test'].params['float_param'].value == 0.1234
    assert puzzle['test'].params['float_param'].get_value() == 0.1234

    # how about one with a format that has fewer decimal points?
    puzzle['test'].params['format_param'].set_value(0.1234)
    assert puzzle['test'].params['format_param'].value == 0.1234
    assert puzzle['test'].params['format_param'].get_value() == 0.1234

    # how about one with an input?
    puzzle['test'].params['input_param'].set_value(0.1234)
    assert puzzle['test'].params['input_param'].value == 0.1234
    assert puzzle['test'].params['input_param'].get_value() == 0.1234

def test_text_param(qtbot, qapp):
    puzzle = pzp.Puzzle(qapp, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    puzzle.show()

    count = [0]
    def count_calls(count=count):
        count[0] += 1
    puzzle['test'].params['text_param'].changed.connect(count_calls)

    assert count[0] == 0

    # The internal method _input_set_value should not result in the Signal being emitted
    # or the internal value changing
    puzzle['test'].params['text_param']._input_set_value('A')
    assert count[0] == 0
    assert puzzle['test'].params['text_param'].value == ""
    assert puzzle['test'].params['text_param'].get_value() == ""

    # In contrast, changing the text field directly (like the user typing)
    # should change the internal value and emit the Signal
    puzzle['test'].params['text_param'].input.setText('B')
    assert count[0] == 1
    assert puzzle['test'].params['text_param'].value == "B"
    assert puzzle['test'].params['text_param'].get_value() == "B"

    # same for changing through the set_value method, which should also update the text field
    puzzle['test'].params['text_param'].set_value('C')
    assert count[0] == 2
    assert puzzle['test'].params['text_param'].value == "C"
    assert puzzle['test'].params['text_param'].get_value() == "C"
    assert puzzle['test'].params['text_param']._input_get_value() == "C"


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    puzzle = pzp.Puzzle(app, "Test params")
    puzzle.add_piece("test", TParamPiece(puzzle), 0, 0)
    from puzzlepiece.pieces import script
    puzzle.add_piece("script", script.Piece(puzzle), 0, 1)
    puzzle.show()
    app.exec()